#![doc = include_str!("../README.md")]
#![warn(clippy::todo)]

use std::{
    fmt::{Debug, Display},
    ops::Range,
};

use range_ext::intersect::{Intersect, IntersectionExt};

/// Operations to apply to a vector
#[derive(Clone)]
pub enum Operation<T> {
    /// No operation
    Nop,
    /// Insert
    ///
    /// Insert n items at the given position.
    ///
    /// **Example:**
    /// ```txt
    /// Ins { pos: 1, val: [4, 5] }
    /// [1,|2, 3] => [1, 4, 5, 2, 3]
    /// ```
    Ins { pos: usize, val: Vec<T> },
    /// Delete
    ///
    /// Delete n items at the given position.
    ///
    /// **Example:**
    /// ```txt
    /// Del { pos: 1, n: 2 }
    /// [1, (2, 3), 4, 5] => [1, 4, 5]
    /// ```
    Del { pos: usize, n: usize },
    /// Move
    ///
    /// Take n items at the given position and move them to another position.
    ///
    /// **Example:**
    /// ```txt
    /// Mov { pos: 0, n: 2, to: 4 }
    /// [(1, 2), 3, 4,|5] => [3, 4, (1, 2), 5]
    /// ```
    Mov { pos: usize, n: usize, to: usize },
    /// Sequence of operations
    Seq { ops: Vec<Operation<T>> },
}

impl<T> Operation<T> {
    /// Get the number of elemets affected by the operation
    ///
    /// # Panics
    ///
    /// Panics if the operation is a [`Operation::Seq`]
    pub fn len(&self) -> usize {
        match self {
            Operation::Nop => 0,
            Operation::Seq { .. } => panic!("SEQ has no len"),
            Operation::Ins { val, .. } => val.len(),
            Operation::Del { n, .. } => *n,
            Operation::Mov { n, .. } => *n,
        }
    }

    /// Get the starting position of the operation
    ///
    /// # Panics
    ///
    /// Panics if the operation is a [`Operation::Nop`] or [`Operation::Seq`]
    pub fn pos(&self) -> usize {
        match self {
            Operation::Nop | Operation::Seq { .. } => panic!("NOP/SEQ has no pos"),
            Operation::Ins { pos, .. }
            | Operation::Del { pos, .. }
            | Operation::Mov { pos, .. } => *pos,
        }
    }

    /// Get the range of the operation
    ///
    /// # Panics
    ///
    /// Panics if the operation is a [`Operation::Nop`] or [`Operation::Seq`]
    pub fn range(&self) -> Range<usize> {
        match self {
            Operation::Nop | Operation::Seq { .. } => panic!("NOP/SEQ has no range"),
            Operation::Ins { pos, val } => *pos..(*pos + val.len()),
            Operation::Del { pos, n } => *pos..(*pos + *n),
            Operation::Mov { pos, n, .. } => *pos..(*pos + *n),
        }
    }

    /// Get the target range of the operation
    ///
    /// # Panics
    ///
    /// Panics if the operation is a [`Operation::Nop`] or [`Operation::Seq`]
    pub fn target_range(&self) -> Range<usize> {
        match self {
            Operation::Nop | Operation::Seq { .. } => panic!("NOP/SEQ has no target"),
            Operation::Ins { pos, val } => *pos..(*pos + val.len()),
            Operation::Del { pos, .. } => *pos..*pos,
            Operation::Mov { to, n, .. } => *to..(*to + *n),
        }
    }

    /// Set the starting position of the operation
    ///
    /// # Panics
    ///
    /// Panics if the operation is a [`Operation::Nop`] or [`Operation::Seq`]
    fn set_pos(&mut self, n_pos: usize) {
        match self {
            Operation::Nop | Operation::Seq { .. } => panic!("cannot move NOP/SEQ"),
            Operation::Ins { pos, .. }
            | Operation::Del { pos, .. }
            | Operation::Mov { pos, .. } => {
                *pos = n_pos;
            }
        }
    }

    /// Move the starting position of the operation by the given number of elements and direction.
    ///
    /// dir=true: add, dir=false: subtract
    ///
    /// Note that this function does not check for overflow
    ///
    /// # Panics
    ///
    /// Panics if the operation is a [`Operation::Nop`] or [`Operation::Seq`]
    fn mov(&mut self, n: usize, dir: bool) {
        match self {
            Operation::Nop | Operation::Seq { .. } => panic!("cannot move nop/seq"),
            Operation::Ins { pos, .. }
            | Operation::Del { pos, .. }
            | Operation::Mov { pos, .. } => {
                if dir {
                    *pos += n;
                } else {
                    *pos -= n;
                }
            }
        }
    }

    /// Apply the operation to the given vector
    ///
    /// # Panics
    ///
    /// Panics if any operation accesses elements that are out of range
    pub fn apply(self, vec: &mut Vec<T>) {
        match self {
            Operation::Nop => {}
            Operation::Ins { pos, val } => {
                splice_or_append(vec, pos, val);
            }
            Operation::Del { pos, n } => {
                vec.splice(pos..(pos + n), None);
            }
            Operation::Mov { pos, n, to } => {
                // If items are moved within the source range, the operation has no effect
                if to > pos && to < pos + n {
                    return;
                }
                let buf = (0..n).map(|_| vec.remove(pos)).collect::<Vec<_>>();
                let n_to = if to > pos { to - n } else { to };
                splice_or_append(vec, n_to, buf);
            }
            Operation::Seq { ops } => ops.into_iter().for_each(|op| op.apply(vec)),
        }
    }

    /// Return true if the operation has no effect
    pub fn is_empty(&self) -> bool {
        match self {
            Operation::Nop => true,
            Operation::Ins { val, .. } => val.is_empty(),
            Operation::Del { n, .. } => *n == 0,
            Operation::Mov { pos, n, to } => {
                *n == 0 || pos == to || (*to >= *pos && *to <= *pos + *n)
            }
            Operation::Seq { ops } => ops.iter().all(Self::is_empty),
        }
    }

    /// Return the 3-character operation name (NOP, INS, DEL, MOV, SEQ)
    fn opname(&self) -> &'static str {
        match self {
            Operation::Nop => "NOP",
            Operation::Ins { .. } => "INS",
            Operation::Del { .. } => "DEL",
            Operation::Mov { .. } => "MOV",
            Operation::Seq { .. } => "SEQ",
        }
    }

    /// Print the operation as it is applied on the given vector
    pub fn print_on(&self, vec: &[T]) -> Vec<T>
    where
        T: Display + Clone,
    {
        if let Operation::Seq { ops } = self {
            let mut v = vec.to_vec();
            for op in ops {
                v = op.print_on(&v)
            }
            return v;
        }

        print!("// {}: [", self.opname());

        if matches!(self, Operation::Nop) {
            for (i, itm) in vec.iter().enumerate() {
                if i > 0 {
                    print!(", ");
                }
                print!("{itm}");
            }
            println!("]");
            return vec.to_vec();
        }

        let r = self.range();
        let r2 = self.target_range();
        let ins = matches!(self, Operation::Ins { .. });
        let del = matches!(self, Operation::Del { .. });
        let mov = matches!(self, Operation::Mov { .. });

        for (i, itm) in vec.iter().enumerate() {
            if i > 0 {
                print!(", ");
            }
            if i == r.start {
                if ins {
                    print!("|");
                } else {
                    print!("(");
                }
            }
            if mov && i == r2.start {
                print!("|");
            }
            print!("{itm}");
            if i == r.end.saturating_sub(1) && !ins {
                print!(")");
            }
        }
        print!("] => [");

        let rt = if mov && r2.start > r.start {
            (r2.start - r2.len())..r2.start
        } else {
            r2.clone()
        };

        let mut v2 = vec.to_vec();
        self.clone().apply(&mut v2);
        for (i, itm) in v2.iter().enumerate() {
            if i > 0 {
                print!(", ");
            }
            if i == rt.start {
                if del {
                    print!("|");
                } else {
                    print!("(");
                }
            }
            print!("{itm}");
            if i == rt.end.saturating_sub(1) && !del {
                print!(")");
            }
        }
        println!("]");
        v2
    }

    fn sanitize(&mut self) {
        if let Operation::Seq { ops } = self {
            ops.iter_mut().for_each(Operation::sanitize);
            for i in (0..ops.len()).rev() {
                if matches!(ops[i], Operation::Nop) {
                    ops.remove(i);
                }
            }
            if ops.is_empty() {
                *self = Operation::Nop;
            } else if ops.len() == 1 {
                *self = ops.pop().unwrap();
            }
            return;
        }
        if self.is_empty() {
            *self = Operation::Nop;
        }
    }
}

impl<T> Debug for Operation<T> {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            Self::Nop => write!(f, "Nop"),
            Self::Ins { pos, val } => f
                .debug_struct("Ins")
                .field("pos", pos)
                .field("val", &val.len())
                .finish(),
            Self::Del { pos, n } => f
                .debug_struct("Del")
                .field("pos", pos)
                .field("n", n)
                .finish(),
            Self::Mov { pos, n, to } => f
                .debug_struct("Mov")
                .field("pos", pos)
                .field("n", n)
                .field("to", to)
                .finish(),
            Self::Seq { ops } => f.debug_struct("Seq").field("ops", ops).finish(),
        }
    }
}

/// Transform the given pair of operations in-place
///
/// Applying the pair of transformed operations a2 and b2 crosswise to the results
/// of a and b leads to a consistent result.
///
/// ```txt
///     /-> a -> b2 -\
/// [V1]              > [V2]
///     \-> b -> a2 -/
/// ```
pub fn transform<T1, T2>(a: &mut Operation<T1>, b: &mut Operation<T2>) {
    transform_internal(a, b);
    a.sanitize();
    b.sanitize();
}

fn transform_internal<T1, T2>(a: &mut Operation<T1>, b: &mut Operation<T2>) {
    a.sanitize();
    b.sanitize();

    match a {
        Operation::Nop => {}
        Operation::Ins { pos: p1, val: v1 } => match b {
            Operation::Nop => {}
            Operation::Ins {
                pos: p2, val: v2, ..
            } => {
                if p1 <= p2 {
                    *p2 += v1.len();
                } else {
                    *p1 += v2.len();
                }
            }
            Operation::Del { .. } => transform_ins_del(a, b),
            Operation::Mov { .. } => transform_mov(b, a, false),
            Operation::Seq { ops } => transform_seq(ops, a, false),
        },
        Operation::Del { pos: p1, n: n1 } => match b {
            Operation::Nop => {}
            Operation::Ins { .. } => transform_ins_del(b, a),
            Operation::Del { pos: p2, n: n2 } => {
                let r1 = *p1..*p1 + *n1;
                let r2 = *p2..*p2 + *n2;

                match r1.intersect_ext(&r2) {
                    IntersectionExt::Empty => {}
                    IntersectionExt::Less => {
                        *p2 -= *n1;
                    }
                    IntersectionExt::LessOverlap => {
                        let overlap = r1.end.saturating_sub(*p2);
                        *n2 -= overlap;
                        *p2 = *p2 + overlap - *n1;
                        *n1 -= overlap;
                    }
                    IntersectionExt::Within => {
                        *n2 -= *n1;
                        *a = Operation::Nop;
                    }
                    IntersectionExt::Same => {
                        *a = Operation::Nop;
                        *b = Operation::Nop;
                    }
                    IntersectionExt::Over => {
                        *n1 -= *n2;
                        *b = Operation::Nop;
                    }
                    IntersectionExt::GreaterOverlap => {
                        let overlap = r2.end.saturating_sub(*p1);
                        *n1 -= overlap;
                        *p1 = *p1 + overlap - *n2;
                        *n2 -= overlap;
                    }
                    IntersectionExt::Greater => {
                        *p1 -= *n2;
                    }
                }
            }
            Operation::Mov { .. } => transform_mov(b, a, false),
            Operation::Seq { ops } => transform_seq(ops, a, false),
        },
        Operation::Mov { .. } => transform_mov(a, b, true),
        Operation::Seq { ops } => transform_seq(ops, b, true),
    }
}

fn transform_ins_del<T1, T2>(ins: &mut Operation<T1>, del: &mut Operation<T2>) {
    let ilen = ins.len();
    let dlen = del.len();

    let ir = ins.range();
    let dr = del.range();

    if ir.start > dr.start && ir.start < dr.end {
        ins.set_pos(dr.start);
        *del = Operation::Seq {
            ops: vec![
                Operation::Del {
                    pos: ir.end,
                    n: dlen - (ir.start - dr.start),
                },
                Operation::Del {
                    pos: dr.start,
                    n: ir.start - dr.start,
                },
            ],
        };
    } else if ir.start <= dr.start {
        del.mov(ilen, true);
    } else {
        ins.set_pos(ir.start.saturating_sub(dlen))
    }
}

fn transform_seq<T1, T2>(a_ops: &mut Vec<Operation<T1>>, b: &mut Operation<T2>, first: bool) {
    if let Operation::Seq { ops: b_ops } = b {
        b_ops
            .iter_mut()
            .for_each(|b_op| transform_seq(a_ops, b_op, first));
    } else {
        a_ops.iter_mut().for_each(|op| {
            if first {
                transform_internal(op, b)
            } else {
                transform_internal(b, op)
            }
        })
    }
}

fn transform_mov<T1, T2>(a: &mut Operation<T1>, b: &mut Operation<T2>, first: bool) {
    let ra = a.range();

    if let Operation::Mov {
        pos: m_pos,
        n: m_n,
        to: m_to,
    } = a
    {
        let m_to_right = m_to > m_pos;
        let m_to_start = if m_to_right { *m_to - *m_n } else { *m_to };
        let rtarget = m_to_start..(m_to_start + *m_n);

        match b {
            Operation::Ins { pos: i_pos, val } => {
                // Is the MOV source split by the insert?
                if *i_pos > ra.start && *i_pos < ra.end {
                    let m_delta = m_to_start as i64 - *m_pos as i64;

                    *i_pos = (*i_pos as i64 + m_delta) as usize;
                    *m_n += val.len();
                    if m_to_right {
                        *m_to += val.len();
                    }
                } else {
                    let mut del = Operation::Del::<()> {
                        pos: *m_pos,
                        n: *m_n,
                    };
                    let mut ins = Operation::Ins::<()> {
                        pos: m_to_start,
                        val: vec![(); *m_n],
                    };

                    if first {
                        transform_internal(&mut del, b);
                        transform_internal(&mut ins, b);
                    } else {
                        transform_internal(b, &mut del);
                        transform_internal(b, &mut ins);
                    }

                    *m_pos = del.pos();
                    *m_to = ins.pos();
                    if m_to_right {
                        *m_to += *m_n;
                    }
                    if m_pos == m_to {
                        *a = Operation::Nop;
                    }
                }
            }
            Operation::Del { pos: d_pos, n: d_n } => {
                let rb = *d_pos..(*d_pos + *d_n);
                let (iba, ibt) = (rb.intersect_ext(&ra), rb.intersect_ext(&rtarget));

                // delete within src - delete within dst
                match &iba {
                    IntersectionExt::Less => {
                        // DEL before MOV
                        if m_to_start <= *d_pos {
                            // moved left from DEL
                            *m_pos -= *d_n;
                            *d_pos += *m_n;
                        } else if matches!(
                            ibt,
                            IntersectionExt::Over | IntersectionExt::LessOverlap
                        ) {
                            // moved inside DEL
                            let split_l = *m_to - *d_pos;
                            *m_pos -= *d_n;
                            *m_to -= split_l;

                            let del_ops = vec![
                                Operation::Del {
                                    pos: rtarget.end,
                                    n: *d_n - split_l,
                                },
                                Operation::Del {
                                    pos: rtarget.start - split_l,
                                    n: split_l,
                                },
                            ];
                            *b = Operation::Seq { ops: del_ops };
                        } else {
                            // moved right from DEL
                            *m_pos -= *d_n;
                            *m_to -= *d_n;
                        }
                    }
                    IntersectionExt::Greater => {
                        // DEL after MOV
                        if matches!(ibt, IntersectionExt::Over | IntersectionExt::GreaterOverlap) {
                            let split_l = *m_to - *d_pos;
                            if split_l == *d_n {
                                *m_to -= *d_n;
                                *d_pos -= *m_n;
                            } else {
                                // moved inside DEL
                                *m_to -= split_l;

                                let del_ops = vec![
                                    Operation::Del {
                                        pos: rtarget.end,
                                        n: *d_n - split_l,
                                    },
                                    Operation::Del {
                                        pos: *d_pos - *m_n,
                                        n: split_l,
                                    },
                                ];
                                *b = Operation::Seq { ops: del_ops };
                            }
                        } else if m_to_start < *d_pos {
                            if !matches!(ibt, IntersectionExt::Greater) {
                                *m_to -= *d_n;
                                *d_pos -= *m_n;
                            }
                        } else {
                            // moved right from DEL
                            *m_to -= *d_n;
                            *d_pos -= *m_n;
                        }
                    }
                    // Part of the source is deleted
                    IntersectionExt::LessOverlap => {
                        // MOV right end of DEL
                        let rest_len = *m_pos - *d_pos;

                        if m_to_right {
                            let del_ops = vec![
                                // Delete moved item
                                Operation::Del {
                                    pos: m_to_start,
                                    n: *d_n - rest_len,
                                },
                                // Delete rest
                                Operation::Del {
                                    pos: *d_pos,
                                    n: rest_len,
                                },
                            ];
                            *m_pos = *d_pos;
                            *m_n = ra.end - rb.end;
                            *m_to -= *d_n;
                            *b = Operation::Seq { ops: del_ops };
                        } else if matches!(
                            &ibt,
                            IntersectionExt::Over | IntersectionExt::LessOverlap
                        ) {
                            /*
                            The DEL range is split by the MOV op in max 3 parts
                            1. DEL before inserted items (d_pos..m_to)
                            2. Inserted items: start may contain deleted items
                               DEL m_to..(m_to + l_intersect)
                            3. DEL after inserted items (m_to+m_n..)
                            */

                            let intersection = ra.start.max(rb.start)..ra.end.min(rb.end);
                            let r_del_mov = *d_pos..(*m_to + intersection.len());

                            let del_ops = vec![
                                Operation::Del {
                                    pos: rtarget.end,
                                    n: *d_n - r_del_mov.len(),
                                },
                                Operation::Del {
                                    pos: r_del_mov.start,
                                    n: r_del_mov.len(),
                                },
                            ];
                            *a = Operation::Nop;
                            *b = Operation::Seq { ops: del_ops };
                        } else {
                            let del_ops = vec![
                                // Delete rest
                                Operation::Del {
                                    pos: *d_pos + *m_n,
                                    n: rest_len,
                                },
                                // Delete moved item
                                Operation::Del {
                                    pos: m_to_start,
                                    n: *d_n - rest_len,
                                },
                            ];
                            *m_pos = *d_pos;
                            *m_n = ra.end - rb.end;
                            *b = Operation::Seq { ops: del_ops };
                        }
                    }
                    IntersectionExt::GreaterOverlap => {
                        // MOV left end of DEL
                        let rest_len = rb.end - ra.end;
                        let moved_len = *d_n - rest_len;
                        let intersection = ra.start.max(rb.start)..ra.end.min(rb.end);

                        if !m_to_right {
                            // MOV to left
                            let del_ops = vec![
                                // Delete rest
                                Operation::Del {
                                    pos: rb.end - (*d_n - intersection.len()),
                                    n: *d_n - intersection.len(),
                                },
                                // Delete moved item
                                Operation::Del {
                                    pos: rtarget.end - intersection.len(),
                                    n: intersection.len(),
                                },
                            ];
                            *m_n -= intersection.len();
                            *b = Operation::Seq { ops: del_ops };
                        } else if matches!(
                            &ibt,
                            IntersectionExt::Over | IntersectionExt::GreaterOverlap
                        ) {
                            let n_right = *m_to - ra.end;

                            let del_ops = vec![
                                Operation::Del {
                                    pos: rtarget.end - intersection.len(),
                                    n: *d_n - n_right,
                                },
                                Operation::Del {
                                    pos: *m_pos,
                                    n: n_right,
                                },
                            ];
                            *a = Operation::Nop;
                            *b = Operation::Seq { ops: del_ops };
                        } else {
                            let del_ops = vec![
                                // Delete moved item
                                Operation::Del {
                                    pos: *m_to - moved_len,
                                    n: moved_len,
                                },
                                // Delete rest
                                Operation::Del {
                                    pos: *m_pos,
                                    n: rest_len,
                                },
                            ];
                            *m_n = *d_pos - *m_pos;
                            *m_to -= *d_n;
                            *b = Operation::Seq { ops: del_ops };
                        }
                    }
                    // DEL middle of the MOV op
                    IntersectionExt::Within => {
                        if m_to_right {
                            let delta = m_to_start - *m_pos;
                            *d_pos += delta;
                            *m_to -= *d_n;
                        } else {
                            let delta = *m_pos - m_to_start;
                            *d_pos -= delta;
                        }
                        *m_n -= *d_n;
                    }
                    // DEL is split by the MOV op
                    IntersectionExt::Over => {
                        let intersection = ra.start.max(rb.start)..ra.end.min(rb.end);
                        if m_to_right {
                            let del_ops = vec![
                                Operation::Del {
                                    pos: rtarget.end - intersection.len(),
                                    n: intersection.len(),
                                },
                                Operation::Del {
                                    pos: *d_pos,
                                    n: *d_n - intersection.len(),
                                },
                            ];
                            *b = Operation::Seq { ops: del_ops };
                        } else {
                            let del_ops = vec![
                                Operation::Del {
                                    pos: rb.end - (*d_n - intersection.len()),
                                    n: *d_n - intersection.len(),
                                },
                                Operation::Del {
                                    pos: (*m_to).min(*d_pos),
                                    n: intersection.len(),
                                },
                            ];
                            *b = Operation::Seq { ops: del_ops };
                        }
                        *a = Operation::Nop;
                    }
                    // Entire source is deleted
                    IntersectionExt::Same => {
                        // Delete all moved items in A'
                        *d_pos = m_to_start;
                        *a = Operation::Nop;
                    }
                    IntersectionExt::Empty => {}
                }
            }
            Operation::Mov {
                pos: m2_pos,
                n: m2_n,
                to: m2_to,
            } => {
                let m2_to_right = m2_to > m2_pos;
                let m2_to_start = if m2_to_right { *m2_to - *m2_n } else { *m2_to };
                let rtarget2 = m2_to_start..(m2_to_start + *m2_n);
                let rm2 = *m2_pos..(*m2_pos + *m2_n);
                let im2 = ra.intersect_ext(&rm2);

                let a_split_by_b = *m2_to > ra.start && *m2_to < ra.end;
                let b_split_by_a = *m_to > rm2.start && *m_to < rm2.end;

                if ra == rm2 {
                    // Asymmetrical, B has priority
                    *m2_pos = m_to_start;
                    *m2_to = m2_to_start;
                    if m2_to > m2_pos {
                        *m2_to += *m_n;
                    }
                    *a = Operation::Nop;
                    return;
                }
                if matches!(im2, IntersectionExt::Within) {
                    // A within B; B has priority, so A should be undone
                    let ops = vec![
                        undo_mov(*m_pos, *m_n, m_to_start),
                        Operation::Mov {
                            pos: *m2_pos,
                            n: *m2_n,
                            to: *m2_to,
                        },
                    ];
                    *b = Operation::Seq { ops };
                    *a = Operation::Nop;
                    return;
                }
                if matches!(im2, IntersectionExt::Over) {
                    // A over B
                    let m2_inside_offset = *m2_pos - *m_pos;
                    let a_before = *m_pos..*m2_pos;
                    let a_after = rm2.end..ra.end;

                    if a_split_by_b {
                        // Split
                        *m2_pos = rtarget.start + m2_inside_offset;
                        if m_to_right {
                            *m2_to += rtarget.start - *m_pos;
                        } else {
                            *m2_to -= *m_pos - rtarget.start;
                        }
                    } else {
                        // A=>B2:
                        let m2_pos_n = rtarget.start + m2_inside_offset;
                        let m2_n_n = *m_n - (a_before.len() + a_after.len());
                        let mut m2_to_n = *m2_to;

                        if *m_pos < *m2_to && rtarget.end >= *m2_to {
                            m2_to_n -= *m_n;
                        } else if *m_pos >= *m2_to && *m_to < *m2_to {
                            m2_to_n += *m_n;
                        }

                        // B=>A2: outer ranges a_before and a_after are lined up
                        // at m_pos if m2_to_right; at m_pos + m2_n if m2_to_left
                        let ab_together_start = if m2_to_right {
                            a_before.start
                        } else {
                            a_before.start + *m2_n
                        };

                        let m_pos_n = ab_together_start;
                        let m_n_n = a_before.len() + a_after.len();
                        let mut m_to_n = *m_to;

                        // Target can be shifted by m2_n to the left if the B operation
                        // source < m_to AND targetB.end >= m_to (B is moved right over m_to)
                        if *m2_pos < *m_to && rtarget2.end >= *m_to {
                            m_to_n -= *m2_n;
                        }
                        // Target shifted by m2_n to the right if the B operation
                        // source >= m_to AND m2_to < m_to (B is moved left over m_to)
                        else if *m2_pos >= *m_to && *m2_to < *m_to {
                            m_to_n += *m2_n;
                        }

                        // If both operations have the same target or split,
                        // the B operation takes priority, A is inserted after it
                        if m_to == m2_to {
                            m2_to_n += *m2_n;
                            if rtarget.contains(&m2_to_n) {
                                m2_to_n = rtarget.end;
                            }
                        }

                        *m_pos = m_pos_n;
                        *m_n = m_n_n;
                        *m_to = m_to_n;
                        *m2_pos = m2_pos_n;
                        *m2_n = m2_n_n;
                        *m2_to = m2_to_n;
                    }
                    return;
                }

                if matches!(im2, IntersectionExt::LessOverlap) {
                    let intersection = *m2_pos..ra.end;
                    match (a_split_by_b, b_split_by_a) {
                        (true, _) => {
                            // Range A is split by B insertion
                            // B: to_left
                            // A=>B2: move intersection of A/B to m_pos
                            let left_of_m2_to = *m_pos..*m2_to;
                            let b_delta = *m2_pos - rtarget2.start;
                            let nleft = *m2_pos - *m_pos;

                            let m2_pos_n = project_mov_point(*m_pos, *m_n, *m_to, *m2_pos);
                            let m2_n_n = intersection.len();
                            let mut m2_to_n = *m_pos;
                            if !m_to_right {
                                m2_to_n += *m_n;
                            }

                            // B=>A2:
                            // 1. move part left of m2_to to m_to
                            // 2. move part right of m2 target to m_to
                            let wi = rm2.contains(m_to); // m_to within m2 source
                            let mut mto = *m_to;
                            if wi {
                                mto -= b_delta;
                            }
                            let a_ops = vec![
                                Operation::Mov {
                                    pos: left_of_m2_to.start,
                                    n: left_of_m2_to.len(),
                                    to: mto,
                                },
                                Operation::Mov {
                                    pos: if wi || !m_to_right {
                                        rtarget2.end
                                    } else {
                                        rtarget2.end - left_of_m2_to.len()
                                    },
                                    n: nleft - left_of_m2_to.len(),
                                    to: if m_to_right {
                                        mto
                                    } else {
                                        mto + left_of_m2_to.len()
                                    },
                                },
                            ];
                            *a = Operation::Seq { ops: a_ops };
                            *m2_pos = m2_pos_n;
                            *m2_n = m2_n_n;
                            *m2_to = m2_to_n;
                            return;
                        }
                        (false, true) => {
                            // Range B is split by A insertion
                            let mut b_ops = None;

                            // A=>B2 move intersection of A/B to m2_to
                            let mdelta = rtarget.start - *m_pos;
                            let m2_pos_n = intersection.start + mdelta;

                            if m2_to < m_pos {
                                let r_before =
                                    (*m2_to + intersection.len())..(*m_pos + intersection.len());
                                b_ops = Some(vec![
                                    Operation::Mov {
                                        pos: m2_pos_n,
                                        n: intersection.len(),
                                        to: *m2_to,
                                    },
                                    Operation::Mov {
                                        pos: r_before.start,
                                        n: r_before.len(),
                                        to: rm2.end,
                                    },
                                ]);
                            }
                            if m2_to_right && m2_to > m_to {
                                let delta = *m2_to - rm2.end; // delta of first op (->)
                                b_ops = Some(vec![
                                    Operation::Mov {
                                        pos: *m_pos,
                                        n: rm2.end - *m_pos,
                                        to: *m2_to,
                                    },
                                    Operation::Mov {
                                        pos: project_mov_point(
                                            *m_pos,
                                            *m_n,
                                            *m_to,
                                            intersection.start,
                                        ) + delta,
                                        n: intersection.len(),
                                        to: *m_pos + delta,
                                    },
                                ]);
                            }

                            // B=>A2 move part left of intersection of A/B to m_to
                            if m2_to_right {
                                let mdelta = rtarget2.start - *m2_pos;
                                *m_to += mdelta;
                            } else {
                                let mdelta = *m2_pos - rtarget2.start;
                                *m_pos += *m2_n;
                                *m_to -= mdelta;
                            }

                            *m_n -= intersection.len();

                            if let Some(ops) = b_ops {
                                *b = Operation::Seq { ops };
                            } else {
                                *m2_n = intersection.len();
                                *m2_pos = m2_pos_n;
                            }
                            return;
                        }
                        (false, false) => {
                            // LessOverlap
                            let intersection = ra.start.max(rm2.start)..ra.end.min(rm2.end);
                            let m_n_n = *m_n - intersection.len();
                            let m_pos_n = project_mov_point(*m2_pos, *m2_n, *m2_to, *m_pos);

                            let mut m_to_n =
                                project_mov_point(*m2_pos, *m2_n, *m2_to, m_to.saturating_sub(1));
                            if *m_to > 0 {
                                m_to_n += 1;
                            }

                            // A=>B2:
                            if m_to == m2_to {
                                *m2_n -= intersection.len();
                                if m_to_right {
                                    *m2_pos = m_pos_n;
                                    *m2_to = *m_to;
                                } else {
                                    *m2_pos = ra.end;
                                    *m2_to = rtarget.end;
                                    m_to_n = *m_to;
                                }
                            } else {
                                // undo A, apply B
                                let b_ops = vec![
                                    undo_mov(*m_pos, *m_n, m_to_start),
                                    Operation::Mov {
                                        pos: *m2_pos,
                                        n: *m2_n,
                                        to: *m2_to,
                                    },
                                    Operation::Mov {
                                        pos: m_pos_n,
                                        n: m_n_n,
                                        to: m_to_n,
                                    },
                                ];
                                *b = Operation::Seq { ops: b_ops };
                            }

                            // B=>A2 mov part left from m2 to m_to
                            *m_pos = m_pos_n;
                            *m_n = m_n_n;
                            *m_to = m_to_n;
                            return;
                        }
                    }
                }

                if matches!(im2, IntersectionExt::GreaterOverlap) {
                    let intersection = *m_pos..rm2.end;
                    match (a_split_by_b, b_split_by_a) {
                        (true, _) => {
                            // Range A is split by B insertion
                            // B: to_right

                            // A=>B2: move intersection of A/B to m_pos
                            let right_of_m2_to = *m2_to..ra.end;
                            let b_delta = rtarget2.start - *m2_pos;
                            let nright = ra.end - rm2.end; // right from overlap

                            let m2_pos_n =
                                project_mov_point(*m_pos, *m_n, *m_to, intersection.start);
                            let m2_n_n = intersection.len();
                            let mut m2_to_n = *m_pos;
                            if !m_to_right {
                                m2_to_n += *m_n;
                            }

                            // B=>A2:
                            // 1. move part right of m2_to to m_to
                            // 2. move part left of m2 target to m_to
                            let wi = rm2.contains(m_to); // m_to within m2 source
                            let mut mto = *m_to;
                            if wi {
                                mto += b_delta;
                            }
                            let prm2_end = project_mov_point(*m2_pos, *m2_n, *m2_to, rm2.end);
                            let a_ops = vec![
                                Operation::Mov {
                                    pos: right_of_m2_to.start,
                                    n: right_of_m2_to.len(),
                                    to: mto,
                                },
                                Operation::Mov {
                                    pos: if m_to_right || wi {
                                        prm2_end
                                    } else {
                                        prm2_end + right_of_m2_to.len()
                                    },
                                    n: nright - right_of_m2_to.len(),
                                    to: if m_to_right {
                                        mto - right_of_m2_to.len()
                                    } else {
                                        mto
                                    },
                                },
                            ];
                            *a = Operation::Seq { ops: a_ops };
                            *m2_pos = m2_pos_n;
                            *m2_n = m2_n_n;
                            *m2_to = m2_to_n;
                            return;
                        }
                        (false, true) => {
                            // Range B is split by A insertion

                            // A=>B2 move intersection of A/B to m2_to
                            let m2_pos_n = rtarget.start;
                            let mut b_ops = None;

                            // B=>A2
                            let r_mov = rm2.end..ra.end;
                            *m_pos = project_mov_point(*m2_pos, *m2_n, *m2_to, r_mov.start);
                            *m_n = r_mov.len();
                            if m2_to_right {
                                let m2delta = rtarget2.start - *m2_pos;
                                *m_to += m2delta;

                                if *m2_to > ra.end {
                                    b_ops = Some(vec![
                                        Operation::Mov {
                                            pos: m2_pos_n,
                                            n: intersection.len(),
                                            to: *m2_to,
                                        },
                                        Operation::Mov {
                                            pos: *m2_pos,
                                            n: ra.end - *m2_pos - intersection.len(),
                                            to: *m2_to - intersection.len(),
                                        },
                                    ]);
                                }
                            } else {
                                let m2delta = *m2_pos - rtarget2.start;
                                *m_to -= m2delta;
                                let d1 = *m2_pos - *m2_to;

                                b_ops = Some(vec![
                                    Operation::Mov {
                                        pos: *m2_pos,
                                        n: ra.end - *m2_pos,
                                        to: *m2_to,
                                    },
                                    Operation::Mov {
                                        pos: m2_pos_n - d1,
                                        n: intersection.len(),
                                        to: ra.end - d1,
                                    },
                                ]);
                            }

                            if let Some(b_ops) = b_ops {
                                *b = Operation::Seq { ops: b_ops }
                            } else {
                                *m2_pos = m2_pos_n;
                                *m2_n = intersection.len();
                            }
                            return;
                        }
                        (false, false) => {
                            // GreaterOverlap
                            let intersection = ra.start.max(rm2.start)..ra.end.min(rm2.end);
                            let m_n_n = *m_n - intersection.len();
                            let m_pos_pre = ra.end - m_n_n;

                            let m_pos_n = project_mov_point(*m2_pos, *m2_n, *m2_to, m_pos_pre);
                            let mut m_to_n = project_mov_point(*m2_pos, *m2_n, *m2_to, *m_to);

                            // A=>B2:
                            if m_to == m2_to {
                                *m2_n -= intersection.len();
                                if m_to_right {
                                    *m2_pos = m_pos_n;
                                    *m2_to = rtarget.start;
                                    m_to_n = *m_to;
                                } else {
                                    *m2_pos = project_mov_point(*m_pos, *m_n, *m_to, rm2.start);
                                    *m2_to = *m_to;
                                }
                            } else {
                                // undo A, apply B
                                let b_ops = vec![
                                    undo_mov(*m_pos, *m_n, m_to_start),
                                    Operation::Mov {
                                        pos: *m2_pos,
                                        n: *m2_n,
                                        to: *m2_to,
                                    },
                                    Operation::Mov {
                                        pos: m_pos_n,
                                        n: m_n_n,
                                        to: m_to_n,
                                    },
                                ];
                                *b = Operation::Seq { ops: b_ops };
                            }

                            // B=>A2 mov part left from m2 to m_to
                            *m_pos = m_pos_n;
                            *m_n = m_n_n;
                            *m_to = m_to_n;

                            return;
                        }
                    }
                }

                if matches!(im2, IntersectionExt::Less) {
                    match (a_split_by_b, b_split_by_a) {
                        (true, false) => {
                            if m_to_right {
                                let mdelta = rtarget.start - *m_pos;

                                // A=>B2: mov m2_src to m2_pos
                                let m2_pos_n = project_mov_point(*m_pos, *m_n, *m_to, *m2_pos);

                                // B=>A2: mov m1_range with inserted m2 items to m_to
                                *m_n += *m2_n;
                                if rm2.end > *m_to {
                                    *m_to += *m2_n;
                                }

                                *m2_pos = m2_pos_n;
                                *m2_to += mdelta;
                            } else {
                                let mdelta = *m_pos - rtarget.start;
                                let m2_pos_n = project_mov_point(*m_pos, *m_n, *m_to, *m2_pos);

                                *m_n += *m2_n;

                                *m2_pos = m2_pos_n;
                                *m2_to -= mdelta;
                            }
                            return;
                        }
                        (false, true) => {
                            if m2_to_right {
                                let m2delta = rtarget2.start - *m2_pos;
                                let m2_pos_n = project_mov_point(*m2_pos, *m2_n, *m2_to, *m_pos);

                                *m2_pos = project_mov_point(*m_pos, *m_n, *m_to, *m2_pos);
                                *m2_n += *m_n;

                                *m_pos = m2_pos_n;
                                *m_to += m2delta;
                            } else {
                                let m2delta = *m2_pos - rtarget2.start;
                                let m2_pos_n = project_mov_point(*m2_pos, *m2_n, *m2_to, *m_pos);

                                *m2_pos = project_mov_point(*m_pos, *m_n, *m_to, *m2_pos);
                                *m2_n += *m_n;
                                if ra.end <= *m2_to {
                                    *m2_to -= *m_n;
                                }

                                *m_pos = m2_pos_n;
                                *m_to -= m2delta;
                            }
                            return;
                        }
                        (true, true) => {
                            // A=>B2: move m2 items (split in 2 ranges) to m2_to
                            let mdelta = rtarget.start - *m_pos;
                            let left_r = *m2_pos..*m_to;
                            let to = *m2_to + mdelta;
                            let b_ops = vec![
                                Operation::Mov {
                                    pos: rtarget.end - *m_n - left_r.len(),
                                    n: left_r.len(),
                                    to,
                                },
                                Operation::Mov {
                                    pos: rtarget.end,
                                    n: *m2_n - left_r.len(),
                                    to,
                                },
                            ];

                            let between = ra.end..rm2.start;
                            if between.is_empty() {
                                *a = Operation::Nop;
                            } else {
                                *m_pos = project_mov_point(*m2_pos, *m2_n, *m2_to, between.start);
                                *m_n = between.len();
                                *m_to = ra.start;
                            }

                            *b = Operation::Seq { ops: b_ops };
                            return;
                        }
                        (false, false) => {}
                    }
                }

                if matches!(im2, IntersectionExt::Greater) {
                    match (a_split_by_b, b_split_by_a) {
                        (true, false) => {
                            if m_to_right {
                                let mdelta = rtarget.start - *m_pos;

                                // A=>B2: mov m2_src to m2_pos
                                let m2_pos_n = project_mov_point(*m_pos, *m_n, *m_to, *m2_pos);

                                // B=>A2: mov m1_range with inserted m2 items to m_to
                                let nl = *m2_to - *m_pos;
                                *m_pos = *m2_to - *m2_n - nl;
                                *m_n += *m2_n;

                                *m2_pos = m2_pos_n;
                                *m2_to += mdelta;
                            } else {
                                let mdelta = *m_pos - rtarget.start;
                                let m2_pos_n = project_mov_point(*m_pos, *m_n, *m_to, *m2_pos);

                                let nl = *m2_to - *m_pos;
                                *m_pos = *m2_to - *m2_n - nl;
                                *m_n += *m2_n;
                                if *m_to > 0 && rm2.end <= *m_to {
                                    *m_to -= *m2_n;
                                }

                                *m2_pos = m2_pos_n;
                                *m2_to -= mdelta;
                            }
                            return;
                        }
                        (false, true) => {
                            if m2_to_right {
                                let m2delta = rtarget2.start - *m2_pos;
                                let m2_pos_n = project_mov_point(*m2_pos, *m2_n, *m2_to, *m_pos);

                                *m2_pos = project_mov_point(*m_pos, *m_n, *m_to, *m2_pos);
                                *m2_n += *m_n;
                                if *m2_to < ra.end {
                                    *m2_to += *m_n;
                                }

                                *m_pos = m2_pos_n;
                                *m_to += m2delta;
                            } else {
                                let m2delta = *m2_pos - rtarget2.start;
                                let m2_pos_n = project_mov_point(*m2_pos, *m2_n, *m2_to, *m_pos);

                                *m2_pos = project_mov_point(*m_pos, *m_n, *m_to, *m2_pos);
                                *m2_n += *m_n;

                                *m_pos = m2_pos_n;
                                *m_to -= m2delta;
                            }
                            return;
                        }
                        (true, true) => {
                            // A=>B2: move m2 items (split in 2 ranges) to m2_to
                            let mdelta = *m_pos - rtarget.start;
                            let left_r = *m2_pos..*m_to;
                            let to = *m2_to - mdelta;
                            let b_ops = vec![
                                Operation::Mov {
                                    pos: rtarget.end - *m_n - left_r.len(),
                                    n: left_r.len(),
                                    to,
                                },
                                Operation::Mov {
                                    pos: rtarget.end,
                                    n: *m2_n - left_r.len(),
                                    to,
                                },
                            ];

                            let between = rm2.end..ra.start;
                            if between.is_empty() {
                                *a = Operation::Nop;
                            } else {
                                *m_pos = project_mov_point(*m2_pos, *m2_n, *m2_to, between.start);
                                *m_n = between.len();
                                *m_to = ra.end;
                            }

                            *b = Operation::Seq { ops: b_ops };
                            return;
                        }
                        (false, false) => {}
                    }
                }

                let mut ia = Operation::<()>::Ins {
                    pos: *m_to,
                    val: vec![(); *m_n],
                };
                let mut da = Operation::<()>::Del {
                    pos: if m_to_right { *m_pos } else { *m_pos + *m_n },
                    n: *m_n,
                };
                let mut ib = Operation::<()>::Ins {
                    pos: *m2_to,
                    val: vec![(); *m2_n],
                };
                let mut db = Operation::<()>::Del {
                    pos: if m2_to_right {
                        *m2_pos
                    } else {
                        *m2_pos + *m2_n
                    },
                    n: *m2_n,
                };

                transform_internal(&mut ia, &mut ib);
                transform_internal(&mut da, &mut ib);
                transform_internal(&mut ia, &mut db);
                transform_internal(&mut da, &mut db);

                *m_pos = if m_to_right {
                    da.pos()
                } else {
                    da.pos() - *m_n
                };
                *m_to = ia.pos();

                *m2_pos = if m2_to_right {
                    db.pos()
                } else {
                    db.pos() - *m2_n
                };
                *m2_to = ib.pos();
            }
            Operation::Seq { ops } => transform_seq(ops, a, false),
            Operation::Nop => {}
        }
    }
}

fn splice_or_append<T>(v1: &mut Vec<T>, pos: usize, mut v2: Vec<T>) {
    if pos < v1.len() {
        v1.splice(pos..pos, v2);
    } else {
        v1.append(&mut v2);
    }
}

fn undo_mov<T>(pos: usize, n: usize, m_to_start: usize) -> Operation<T> {
    let mop1_to = if pos > m_to_start { pos + n } else { pos };
    Operation::Mov {
        pos: m_to_start,
        n,
        to: mop1_to,
    }
}

fn project_mov_point(m_pos: usize, m_n: usize, m_to: usize, point: usize) -> usize {
    let m_to_right = m_to > m_pos;
    let m_to_start = if m_to_right { m_to - m_n } else { m_to };
    let rmov = m_pos..(m_pos + m_n);
    let rtarget = m_to_start..(m_to_start + m_n);
    let m_delta = m_to_start as i64 - m_pos as i64;

    if (point < m_pos && point < rtarget.start) || (point >= rmov.end && point >= m_to) {
        // unaffected
        point
    } else if rmov.contains(&point) {
        // point is moved
        (point as i64 + m_delta) as usize
    } else if m_to_right {
        point - m_n
    } else {
        point + m_n
    }
}

#[cfg(test)]
mod tests {
    use super::*;

    fn filter_to(size: usize, pos: usize, to: usize) -> usize {
        if pos == to {
            if to >= size - 1 {
                filter_to(size, pos, 0)
            } else {
                filter_to(size, pos, to + 1)
            }
        } else {
            to
        }
    }

    #[test]
    fn t_project_mov_point() {
        let size = 9;
        let input = [0, 1, 2, 3, 4, 5, 6, 7, 8];

        (0..size).into_iter().for_each(|m_pos| {
            (1..(size - m_pos)).into_iter().for_each(|m_n| {
                (0..size).into_iter().for_each(|m_to| {
                    (0..size).into_iter().for_each(|point| {
                        let m_to = filter_to(size, m_pos, m_to);
                        let m_n = m_n.min(size - m_pos.max(m_to));

                        let mov = Operation::Mov {
                            pos: m_pos,
                            n: m_n,
                            to: m_to,
                        };
                        if mov.is_empty() {
                            return;
                        }

                        let res = project_mov_point(m_pos, m_n, m_to, point);

                        let mut data = input.to_vec();
                        mov.clone().apply(&mut data);
                        let moved_to = data
                            .iter()
                            .enumerate()
                            .find(|(_, itm)| **itm == point as i32)
                            .unwrap()
                            .0;

                        if res != moved_to {
                            mov.print_on(&input);
                            println!("Point: {point}");
                            assert_eq!(res, moved_to);
                        }
                    })
                })
            })
        });
    }
}
